package uk.ac.starlink.votable.tools;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import org.xml.sax.SAXException;
import uk.ac.starlink.table.StoragePolicy;
import uk.ac.starlink.util.DataSource;
import uk.ac.starlink.votable.DataFormat;
import uk.ac.starlink.votable.VOElementFactory;

/**
 * Copies a VOTable XML document intact but with control over the 
 * DATA encoding type.
 * Currently, lexical details such as ignorable whitespace and comments
 * may be mangled, and namespace information may be lost.
 *
 * @author   Mark Taylor (Starlink)
 * @since    22 Sep 2004
 */
public class Copy {

    /**
     * Main method.  See usage message (-h) for details.
     */
    public static void main( String[] args ) {
        String cmdname = "votcopy";
        String usage = cmdname 
                     + " [-h[elp]]"
                     + " [-disk]"
                     + " [-base name]"
                     + " [-f[ormat] tabledata|binary|fits|none]"
                     + " [<in> [<out>]]";

        /* Process flags. */
        List argList = new ArrayList( Arrays.asList( args ) );
        DataFormat format = DataFormat.TABLEDATA;
        File base = null;
        StoragePolicy storagePolicy = StoragePolicy.getDefaultPolicy();
        for ( Iterator it = argList.iterator(); it.hasNext(); ) {
            String arg = (String) it.next();
            if ( arg.equals( "-f" ) || arg.equals( "-format" ) ) {
                it.remove();
                if ( it.hasNext() ) {
                    String fname = (String) it.next();
                    it.remove();
                    if ( fname.equalsIgnoreCase( "tabledata" ) ) {
                        format = DataFormat.TABLEDATA;
                    }
                    else if ( fname.equalsIgnoreCase( "fits" ) ) {
                        format = DataFormat.FITS;
                    }
                    else if ( fname.equalsIgnoreCase( "binary" ) ) {
                        format = DataFormat.BINARY;
                    }
                    else if ( fname.equalsIgnoreCase( "none" ) ) {
                        format = null;
                    }
                    else {
                        System.err.println( usage );
                        System.exit( 1 );
                    }
                }
            }
            else if ( arg.equals( "-base" ) ) {
                it.remove();
                if ( it.hasNext() && base == null ) {
                    base = new File( (String) it.next() );
                    it.remove();
                }
                else {
                    System.err.println( usage );
                    System.exit( 1 );
                }
            }
            else if ( arg.equals( "-disk" ) ) {
                it.remove();
                storagePolicy = StoragePolicy.PREFER_DISK;
            }
            else if ( arg.equals( "-h" ) || arg.equals( "-help" ) ) {
                System.out.println( usage );
                System.exit( 0 );
            }
            else if ( arg.startsWith( "-" ) && arg.length() > 1 ) {
                System.err.println( usage );
                System.exit( 1 );
            }
        }

        /* Get non-flag arguments. */
        String inName = argList.isEmpty() ? "-" : (String) argList.remove( 0 );
        String outName = argList.isEmpty() ? "-" : (String) argList.remove( 0 );
        if ( ! argList.isEmpty() ) {
            System.err.println( usage );
            System.exit( 1 );
        }

        /* Get input and output streams. */
        String systemId = inName.equals( "-" ) ? "." : inName;
        InputStream in = null;
        OutputStream out = null;
        try {
            in = DataSource.getInputStream( inName );
            out = outName.equals( "-" ) 
                ? (OutputStream) System.out
                : (OutputStream) new FileOutputStream( outName );
            in = new BufferedInputStream( in );
            out = new BufferedOutputStream( out );

            /* Do the copy. */
            copyVOTable( systemId, in, out, format, storagePolicy, base );
            out.flush();
        }
        catch ( IOException e ) {
            System.err.println( e.toString() );
            System.exit( 1 );
        }
        catch ( SAXException e ) {
            System.err.println( e.toString() );
            System.exit( 1 );
        }
        finally {
            try {
                if ( in != null ) {
                    in.close();
                }
                if ( out != null ) {
                    out.close();
                }
            }
            catch ( IOException e ) {
                // no action
            }
        }
    }

    /**
     * Copies a VOTable from one stream to another, outputting the
     * data in a given format.
     * If the <tt>base</tt> parameter has a non-null value, it is used
     * as the base filename for href-referenced FITS/BINARY data files.
     * If it's <tt>null</tt>, then such data will be base64-encoded
     * in-line.
     *
     * @param   systemId  system ID for input stream, or null
     * @param   in     input stream
     * @param   out    output stream
     * @param   format VOTable output variant
     * @param   policy  storage policy if data needs to be stored
     * @param   base   filename which forms base for href-referenced 
     *                 files
     */
    public static void copyVOTable( String systemId, InputStream in,
                                    OutputStream out, DataFormat format,
                                    StoragePolicy policy, File base )
            throws IOException, SAXException {

        /* Don't store rows if we're never going to output them. */
        if ( format == null ) {
            policy = StoragePolicy.DISCARD;
        }

        /* Parse the input to a reduced DOM. */
        DOMSource dsrc = new VOElementFactory( policy )
                        .transformToDOM( new StreamSource( in, systemId ),
                                         false );
        DOMWriter writer = new DOMWriter( out, format, base );
        try {
            writer.writeNode( dsrc.getNode() );
        }
        finally {
            writer.close();
        }
    }
}
